"""
This module abstracts different XML flavors used to represent the underlying invoice data.

Data kept for each flavor:

- mapping between "pivot" dict and XML paths
- xmp templates
- xsd files for validation
- xml templates to create new XML representations
"""

import os
import yaml
from lxml import etree
from io import BytesIO
from pkg_resources import resource_filename

from ..logger import logger

unicode = str

# Load information on different XML standards and paths from YML.
def _load_yml(filename):
    with open(os.path.join(os.path.dirname(__file__), filename)) as f:
        return yaml.load(f)

FIELDS = _load_yml('fields.yml')
FLAVORS = _load_yml('flavors.yml')


class XMLFlavor(object):
    """A helper class to keep the lookup code out of the main library.

    Represents a XML invoice representation standard, like Factur-X or Zugferd.
    """

    def __init__(self, xml):
        self.name = guess_flavor(xml)
        self.level = self.get_level(xml)
        self.details = FLAVORS[self.name]

    @classmethod
    def from_template(cls, flavor, level):
        """Creates a new XML tree with the desired level and flavor from an existing template

        Returns lxml.etree and xml_flavor.XMLFlavor instance.
        """
        template_filename = os.path.join(
            os.path.dirname(__file__),
            flavor, 
            'xml',
             FLAVORS[flavor]['levels'][level]['xml'])
        assert os.path.isfile(template_filename), 'Template for this flavor/level does not exist.'
        
        xml_tree = etree.parse(open(template_filename)).getroot()
        return cls(xml_tree), xml_tree

    def get_level(self, facturx_xml_etree):
        if not isinstance(facturx_xml_etree, type(etree.Element('pouet'))):
            raise ValueError('facturx_xml_etree must be an etree.Element() object')
        namespaces = facturx_xml_etree.nsmap
        doc_id_xpath = facturx_xml_etree.xpath(self._get_xml_path('version'), namespaces=namespaces)
        if not doc_id_xpath:
            raise ValueError("Version field not found.")
        doc_id = doc_id_xpath[0].text
        level = doc_id.split(':')[-1]
        if level not in FLAVORS[self.name]['levels']:
            level = doc_id.split(':')[-2]
        if level not in FLAVORS[self.name]['levels']:
            raise ValueError(
                "Invalid Factur-X URN: '%s'" % doc_id)
        logger.info('Factur-X level is %s (autodetected)', level)
        return level


    def check_xsd(self, etree_to_validate):
        """Validate the XML file against the XSD"""

        xsd_filename = FLAVORS[self.name]['levels'][self.level]['schema']
        xsd_file = os.path.join(
            os.path.dirname(__file__),
            self.name, 'xsd', xsd_filename)

        xsd_etree_obj = etree.parse(open(xsd_file))
        official_schema = etree.XMLSchema(xsd_etree_obj)
        try:
            official_schema.assertValid(etree_to_validate)
            logger.info('XML file successfully validated against XSD')
        except Exception as e:
            # if the validation of the XSD fails, we arrive here
            logger.error(
                "The XML file is invalid against the XML Schema Definition")
            logger.error('XSD Error: %s', e)
            raise Exception(
                "The %s XML file is not valid against the official "
                "XML Schema Definition. "
                "Here is the error, which may give you an idea on the "
                "cause of the problem: %s." % (self.name, unicode(e)))
        return True

    def get_xmp_xml(self):
        xmp_file = os.path.join(
            os.path.dirname(__file__),
            self.name, 
            'xmp',
             FLAVORS[self.name]['xmp_schema'])
        return etree.parse(open(xmp_file))

    def _get_xml_path(self, field_name):
        """Return XML path based on field_name and flavor"""

        assert field_name in FIELDS.keys(), 'Field not specified. Try working directly on the XML tree.'
        field_details = FIELDS[field_name]
        if self.name in field_details['_path']:
            return field_details['_path'][self.name]
        else:
            raise KeyError('Path not defined for currenct flavor.')

def valid_xmp_filenames():
    result = []
    for flavor in FLAVORS.keys():
        result.append(FLAVORS[flavor]['xmp_filename'])
    return result

def guess_flavor(facturx_xml_etree):
    if not isinstance(facturx_xml_etree, type(etree.Element('pouet'))):
        raise ValueError('facturx_xml_etree must be an etree.Element() object')
    if facturx_xml_etree.tag.startswith('{urn:un:unece:uncefact:'):
        flavor = 'factur-x'
    elif facturx_xml_etree.tag.startswith('{urn:ferd:'):
        flavor = 'zugferd'
    else:
        raise Exception(
            "Could not detect if the invoice is a Factur-X or ZUGFeRD "
            "invoice.")
    logger.info('Factur-X flavor is %s (autodetected)', flavor)
    return flavor



import io
import os
import yaml
import codecs
from io import BytesIO
from lxml import etree
from tempfile import NamedTemporaryFile
from datetime import datetime
import os.path
import mimetypes
import hashlib
from PyPDF2 import PdfFileReader
from PyPDF2.generic import IndirectObject
import json

from .flavors import xml_flavor
from .logger import logger
from .pdfwriter import FacturXPDFWriter

# Python 2 and 3 compat
try:
    file_types = (file, io.IOBase)
except NameError:
    file_types = (io.IOBase,)
unicode = str

__all__ = ['FacturX']


class FacturX(object):
    """Represents an electronic PDF invoice with embedded XML metadata following the
    Factur-X standard.

    The source of truth is always the underlying XML tree. No copy of field
    data is kept. Manipulation of the XML tree is either done via Python-style
    dict access (available for the most common fields) or by directly accessing
    the XML data on `FacturX.xml`.

    Attributes:
    - xml: xml tree of machine-readable representation.
    - pdf: underlying graphical PDF representation.
    - flavor: which flavor (Factur-x or Zugferd) to use.
    """

    def __init__(self, pdf_invoice, flavor='factur-x', level='minimum'):
        # Read PDF from path, pointer or string
        if isinstance(pdf_invoice, str) and pdf_invoice.endswith('.pdf') and os.path.isfile(pdf_invoice):
            with open(pdf_invoice, 'rb') as f:
                pdf_file = BytesIO(f.read())
        elif isinstance(pdf_invoice, str):
            pdf_file = BytesIO(pdf_invoice)
        elif isinstance(pdf_invoice, file_types):
            pdf_file = pdf_invoice
        else:
            raise TypeError(
                "The first argument of the method get_facturx_xml_from_pdf must "
                "be either a string or a file (it is a %s)." % type(pdf_invoice))

        xml = self._xml_from_file(pdf_file)
        self.pdf = pdf_file

        # PDF has metadata embedded
        if xml is not None:
            self.xml = xml
            self.flavor = xml_flavor.XMLFlavor(xml)
            logger.info('Read existing XML from PDF. Flavor: %s', self.flavor.name)
        # No metadata embedded. Create from template.
        else:
            self.flavor, self.xml = xml_flavor.XMLFlavor.from_template(flavor, level)
            logger.info('PDF does not have XML embedded. Adding from template.')

        self.flavor.check_xsd(self.xml)
        self._namespaces = self.xml.nsmap

    def read_xml(self):
        """Use XML data from external file. Replaces existing XML or template."""
        pass

    def _xml_from_file(self, pdf_file):
        pdf = PdfFileReader(pdf_file)
        pdf_root = pdf.trailer['/Root']
        if '/Names' not in pdf_root or '/EmbeddedFiles' not in pdf_root['/Names']:
            logger.info('No existing XML file found.')
            return None

        for file in pdf_root['/Names']['/EmbeddedFiles']['/Names']:
            if isinstance(file, IndirectObject):
                obj = file.getObject()
                if obj['/F'] in xml_flavor.valid_xmp_filenames():
                    xml_root = etree.fromstring(obj['/EF']['/F'].getData())
                    xml_content = xml_root
                    xml_filename = obj['/F']
                    logger.info(
                        'A valid XML file %s has been found in the PDF file',
                        xml_filename)
                    return xml_content

    def __getitem__(self, field_name):
        path = self.flavor._get_xml_path(field_name)
        value = self.xml.xpath(path, namespaces=self._namespaces)
        if value is not None:
            value = value[0].text
        if 'date' in field_name:
            value = datetime.strptime(value, '%Y%m%d')
        return value

    def __setitem__(self, field_name, value):
        path = self.flavor._get_xml_path(field_name)
        res = self.xml.xpath(path, namespaces=self._namespaces)
        if len(res) > 1:
            raise LookupError('Multiple nodes found for this path. Refusing to edit.')

        if 'date' in field_name:
            assert isinstance(value, datetime), 'Please pass date values as DateTime() object.'
            value = value.strftime('%Y%m%d')
            res[0].attrib['format'] = '102'
            res[0].text = value
        else:
            res[0].text = value

    def is_valid(self):
        """Make every effort to validate the current XML.

        Checks:
        - all required fields are present and have values.
        - XML is valid
        - ...

        Returns: true/false (validation passed/failed)
        """
        # validate against XSD
        try:
            self.flavor.check_xsd(self.xml)
        except Exception:
            return False

        # Check for required fields
        fields_data = xml_flavor.FIELDS
        for field in fields_data.keys():
            if fields_data[field]['_required']:
                r = self.xml.xpath(fields_data[field]['_path'][self.flavor.name], namespaces=self._namespaces)
                if not len(r):
                    logger.error("Required field '%s' is not present", field)
                    return False
                elif r[0].text is None:
                    logger.error("Required field %s doesn't contain any value", field)
                    return False

        return True
    
    def write_pdf(self, path):
        pdfwriter = FacturXPDFWriter(self)
        with open(path, 'wb') as output_f:
            pdfwriter.write(output_f)

        logger.info('XML file added to PDF invoice')
        return True

    @property
    def xml_str(self):
        """Calculate MD5 checksum of XML file. Used for PDF attachment."""
        return etree.tostring(self.xml, pretty_print=True)

    def write_xml(self, path):
        with open(path, 'wb') as f:
            f.write(self.xml_str)

    def __make_dict(self):
        fields_data = xml_flavor.FIELDS
        flavor = self.flavor.name

        output_dict = {}
        for field in fields_data.keys():
            try:
                r = self.xml.xpath(fields_data[field]['_path'][flavor], namespaces=self._namespaces)
                output_dict[field] = r[0].text
            except IndexError:
                output_dict[field] = None

        return output_dict

    def write_json(self, json_file_path='output.json'):
        json_output = self.__make_dict()
        # if self.is_valid():
        #     with open(json_file_path, 'w') as json_file:
        #         logger.info("Exporting JSON to %s", json_file_path)
        #         json.dump(json_output, json_file, indent=4, sort_keys=True)

        with open(json_file_path, 'w') as json_file:
                logger.info("Exporting JSON to %s", json_file_path)
                json.dump(json_output, json_file, indent=4, sort_keys=True)

    def write_yml(self, yml_file_path='output.yml'):
        yml_output = self.__make_dict()
        # if self.is_valid():
        #     with open(yml_file_path, 'w') as yml_file:
        #         logger.info("Exporting YAML to %s", yml_file_path)
        #         yaml.dump(yml_output, yml_file, default_flow_style=False)

        with open(yml_file_path, 'w') as yml_file:
                logger.info("Exporting YAML to %s", yml_file_path)
                yaml.dump(yml_output, yml_file, default_flow_style=False)



# -*- coding: utf-8 -*-
# Copyright 2016-2018, Alexis de Lattre <alexis.delattre@akretion.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * The name of the authors may not be used to endorse or promote products
#      derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# TODO list:
# - have both python2 and python3 support
# - add automated tests (currently, we only have tests at odoo module level)
# - keep original metadata by copy of pdf_tailer[/Info] ?

from ._version import __version__
import io
import os
import yaml
import codecs
from io import BytesIO
from lxml import etree
from tempfile import NamedTemporaryFile
from datetime import datetime
from PyPDF2 import PdfFileWriter, PdfFileReader
from PyPDF2.generic import DictionaryObject, DecodedStreamObject,\
    NameObject, createStringObject, ArrayObject, IndirectObject
from pkg_resources import resource_filename
import os.path
import mimetypes
import hashlib

from .logger import logger

# Python 2 and 3 compat
try:
    file_types = (file, io.IOBase)
except NameError:
    file_types = (io.IOBase,)
unicode = str

class FacturXPDFWriter(PdfFileWriter):
    def __init__(self, facturx, pdf_metadata=None):
        """Take a FacturX instance and write the XML to the attached PDF file"""

        super(FacturXPDFWriter, self).__init__()
        # TODO: Can handle str/paths and ByteIO?
        self.factx = facturx

        original_pdf = PdfFileReader(facturx.pdf)
        # Extract /OutputIntents obj from original invoice
        output_intents = _get_original_output_intents(original_pdf)
        self.appendPagesFromReader(original_pdf)

        original_pdf_id = original_pdf.trailer.get('/ID')
        logger.debug('original_pdf_id=%s', original_pdf_id)
        if original_pdf_id:
            self._ID = original_pdf_id
            # else : generate some ?

        if pdf_metadata is None:
            base_info = {
                'seller': self.factx['seller'],
                'number': self.factx['invoice_number'],
                'date': self.factx['date'],
                'doc_type': self.factx['type'],
                }
            pdf_metadata = _base_info2pdf_metadata(base_info)
        else:
            # clean-up pdf_metadata dict
            for key, value in pdf_metadata.iteritems():
                if not isinstance(value, (str, unicode)):
                    pdf_metadata[key] = ''

        self._update_metadata_add_attachment(pdf_metadata, output_intents)


    def _update_metadata_add_attachment(self, pdf_metadata, output_intents):
        '''This method is inspired from the code of the addAttachment()
        method of the PyPDF2 lib'''
        
        # The entry for the file
        facturx_xml_str = self.factx.xml_str
        md5sum = hashlib.md5().hexdigest()
        md5sum_obj = createStringObject(md5sum)
        params_dict = DictionaryObject({
            NameObject('/CheckSum'): md5sum_obj,
            NameObject('/ModDate'): createStringObject(_get_pdf_timestamp()),
            NameObject('/Size'): NameObject(str(len(facturx_xml_str))),
            })
        file_entry = DecodedStreamObject()
        file_entry.setData(facturx_xml_str)  # here we integrate the file itself
        file_entry.update({
            NameObject("/Type"): NameObject("/EmbeddedFile"),
            NameObject("/Params"): params_dict,
            # 2F is '/' in hexadecimal
            NameObject("/Subtype"): NameObject("/text#2Fxml"),
            })
        file_entry_obj = self._addObject(file_entry)
        # The Filespec entry
        ef_dict = DictionaryObject({
            NameObject("/F"): file_entry_obj,
            NameObject('/UF'): file_entry_obj,
            })

        xmp_filename = self.factx.flavor.details['xmp_filename']
        fname_obj = createStringObject(xmp_filename)
        filespec_dict = DictionaryObject({
            NameObject("/AFRelationship"): NameObject("/Data"),
            NameObject("/Desc"): createStringObject("Factur-X Invoice"),
            NameObject("/Type"): NameObject("/Filespec"),
            NameObject("/F"): fname_obj,
            NameObject("/EF"): ef_dict,
            NameObject("/UF"): fname_obj,
            })
        filespec_obj = self._addObject(filespec_dict)
        name_arrayobj_cdict = {fname_obj: filespec_obj}
        
        # TODO: add back additional attachments?
        logger.debug('name_arrayobj_cdict=%s', name_arrayobj_cdict)
        name_arrayobj_content_sort = list(
            sorted(name_arrayobj_cdict.items(), key=lambda x: x[0]))
        logger.debug('name_arrayobj_content_sort=%s', name_arrayobj_content_sort)
        name_arrayobj_content_final = []
        af_list = []
        for (fname_obj, filespec_obj) in name_arrayobj_content_sort:
            name_arrayobj_content_final += [fname_obj, filespec_obj]
            af_list.append(filespec_obj)
        embedded_files_names_dict = DictionaryObject({
            NameObject("/Names"): ArrayObject(name_arrayobj_content_final),
            })
        
        # Then create the entry for the root, as it needs a
        # reference to the Filespec
        embedded_files_dict = DictionaryObject({
            NameObject("/EmbeddedFiles"): embedded_files_names_dict,
            })
        res_output_intents = []
        logger.debug('output_intents=%s', output_intents)
        for output_intent_dict, dest_output_profile_dict in output_intents:
            dest_output_profile_obj = self._addObject(
                dest_output_profile_dict)
            # TODO detect if there are no other objects in output_intent_dest_obj
            # than /DestOutputProfile
            output_intent_dict.update({
                NameObject("/DestOutputProfile"): dest_output_profile_obj,
                })
            output_intent_obj = self._addObject(output_intent_dict)
            res_output_intents.append(output_intent_obj)
        
        # Update the root
        xmp_level_str = self.factx.flavor.details['levels'][self.factx.flavor.level]['xmp_str']
        xmp_template = self.factx.flavor.get_xmp_xml()
        metadata_xml_str = _prepare_pdf_metadata_xml(xmp_level_str, xmp_filename, xmp_template, pdf_metadata)
        metadata_file_entry = DecodedStreamObject()
        metadata_file_entry.setData(metadata_xml_str)
        metadata_file_entry.update({
            NameObject('/Subtype'): NameObject('/XML'),
            NameObject('/Type'): NameObject('/Metadata'),
            })
        metadata_obj = self._addObject(metadata_file_entry)
        af_value_obj = self._addObject(ArrayObject(af_list))
        self._root_object.update({
            NameObject("/AF"): af_value_obj,
            NameObject("/Metadata"): metadata_obj,
            NameObject("/Names"): embedded_files_dict,
            # show attachments when opening PDF
            NameObject("/PageMode"): NameObject("/UseAttachments"),
            })
        logger.debug('res_output_intents=%s', res_output_intents)
        if res_output_intents:
            self._root_object.update({
                NameObject("/OutputIntents"): ArrayObject(res_output_intents),
            })
        metadata_txt_dict = _prepare_pdf_metadata_txt(pdf_metadata)
        self.addMetadata(metadata_txt_dict)



def _get_metadata_timestamp():
    now_dt = datetime.now()
    # example format : 2014-07-25T14:01:22+02:00
    meta_date = now_dt.strftime('%Y-%m-%dT%H:%M:%S+00:00')
    return meta_date

def _base_info2pdf_metadata(base_info):
    if base_info['doc_type'] == '381':
        doc_type_name = u'Refund'
    else:
        doc_type_name = u'Invoice'
    date_str = datetime.strftime(base_info['date'], '%Y-%m-%d')
    title = '%s: %s %s' % (
        base_info['seller'], doc_type_name, base_info['number'])
    subject = 'Factur-X %s %s dated %s issued by %s' % (
        doc_type_name, base_info['number'], date_str, base_info['seller'])
    pdf_metadata = {
        'author': base_info['seller'],
        'keywords': u'%s, Factur-X' % doc_type_name,
        'title': title,
        'subject': subject,
        }
    logger.debug('Converted base_info to pdf_metadata: %s', pdf_metadata)
    return pdf_metadata


def _prepare_pdf_metadata_txt(pdf_metadata):
    pdf_date = _get_pdf_timestamp()
    info_dict = {
        '/Author': pdf_metadata.get('author', ''),
        '/CreationDate': pdf_date,
        '/Creator':
        u'factur-x Python lib v%s by Alexis de Lattre' % __version__,
        '/Keywords': pdf_metadata.get('keywords', ''),
        '/ModDate': pdf_date,
        '/Subject': pdf_metadata.get('subject', ''),
        '/Title': pdf_metadata.get('title', ''),
        }
    return info_dict


def _prepare_pdf_metadata_xml(xmp_level_str, xmp_filename, facturx_ext_schema_root, pdf_metadata):
    nsmap_x = {'x': 'adobe:ns:meta/'}
    nsmap_rdf = {'rdf': 'http://www.w3.org/1999/02/22-rdf-syntax-ns#'}
    nsmap_dc = {'dc': 'http://purl.org/dc/elements/1.1/'}
    nsmap_pdf = {'pdf': 'http://ns.adobe.com/pdf/1.3/'}
    nsmap_xmp = {'xmp': 'http://ns.adobe.com/xap/1.0/'}
    nsmap_pdfaid = {'pdfaid': 'http://www.aiim.org/pdfa/ns/id/'}
    nsmap_fx = {
        'fx': 'urn:factur-x:pdfa:CrossIndustryDocument:invoice:1p0#'}
    ns_x = '{%s}' % nsmap_x['x']
    ns_dc = '{%s}' % nsmap_dc['dc']
    ns_rdf = '{%s}' % nsmap_rdf['rdf']
    ns_pdf = '{%s}' % nsmap_pdf['pdf']
    ns_xmp = '{%s}' % nsmap_xmp['xmp']
    ns_pdfaid = '{%s}' % nsmap_pdfaid['pdfaid']
    ns_fx = '{%s}' % nsmap_fx['fx']
    ns_xml = '{http://www.w3.org/XML/1998/namespace}'

    root = etree.Element(ns_x + 'xmpmeta', nsmap=nsmap_x)
    rdf = etree.SubElement(
        root, ns_rdf + 'RDF', nsmap=nsmap_rdf)
    desc_pdfaid = etree.SubElement(
        rdf, ns_rdf + 'Description', nsmap=nsmap_pdfaid)
    desc_pdfaid.set(ns_rdf + 'about', '')
    etree.SubElement(
        desc_pdfaid, ns_pdfaid + 'part').text = '3'
    etree.SubElement(
        desc_pdfaid, ns_pdfaid + 'conformance').text = 'B'
    desc_dc = etree.SubElement(
        rdf, ns_rdf + 'Description', nsmap=nsmap_dc)
    desc_dc.set(ns_rdf + 'about', '')
    dc_title = etree.SubElement(desc_dc, ns_dc + 'title')
    dc_title_alt = etree.SubElement(dc_title, ns_rdf + 'Alt')
    dc_title_alt_li = etree.SubElement(
        dc_title_alt, ns_rdf + 'li')
    dc_title_alt_li.text = pdf_metadata.get('title', '')
    dc_title_alt_li.set(ns_xml + 'lang', 'x-default')
    dc_creator = etree.SubElement(desc_dc, ns_dc + 'creator')
    dc_creator_seq = etree.SubElement(dc_creator, ns_rdf + 'Seq')
    etree.SubElement(
        dc_creator_seq, ns_rdf + 'li').text = pdf_metadata.get('author', '')
    dc_desc = etree.SubElement(desc_dc, ns_dc + 'description')
    dc_desc_alt = etree.SubElement(dc_desc, ns_rdf + 'Alt')
    dc_desc_alt_li = etree.SubElement(
        dc_desc_alt, ns_rdf + 'li')
    dc_desc_alt_li.text = pdf_metadata.get('subject', '')
    dc_desc_alt_li.set(ns_xml + 'lang', 'x-default')
    desc_adobe = etree.SubElement(
        rdf, ns_rdf + 'Description', nsmap=nsmap_pdf)
    desc_adobe.set(ns_rdf + 'about', '')
    producer = etree.SubElement(
        desc_adobe, ns_pdf + 'Producer')
    producer.text = 'PyPDF2'
    desc_xmp = etree.SubElement(
        rdf, ns_rdf + 'Description', nsmap=nsmap_xmp)
    desc_xmp.set(ns_rdf + 'about', '')
    creator = etree.SubElement(
        desc_xmp, ns_xmp + 'CreatorTool')
    creator.text = 'factur-x python lib v%s by Alexis de Lattre' % __version__
    timestamp = _get_metadata_timestamp()
    etree.SubElement(desc_xmp, ns_xmp + 'CreateDate').text = timestamp
    etree.SubElement(desc_xmp, ns_xmp + 'ModifyDate').text = timestamp

    # The Factur-X extension schema must be embedded into each PDF document
    facturx_ext_schema_desc_xpath = facturx_ext_schema_root.xpath(
        '//rdf:Description', namespaces=nsmap_rdf)
    rdf.append(facturx_ext_schema_desc_xpath[1])
    # Now is the Factur-X description tag
    facturx_desc = etree.SubElement(
        rdf, ns_rdf + 'Description', nsmap=nsmap_fx)
    facturx_desc.set(ns_rdf + 'about', '')
    facturx_desc.set(
        ns_fx + 'ConformanceLevel', xmp_level_str)
    facturx_desc.set(ns_fx + 'DocumentFileName', xmp_filename)
    facturx_desc.set(ns_fx + 'DocumentType', 'INVOICE')
    facturx_desc.set(ns_fx + 'Version', '1.0')

    # TODO: should be UTF-16be ??
    xml_str = etree.tostring(
        root, pretty_print=True, encoding="UTF-8", xml_declaration=False)
    head = u'<?xpacket begin="\ufeff" id="W5M0MpCehiHzreSzNTczkc9d"?>'.encode(
        'utf-8')
    tail = u'<?xpacket end="w"?>'.encode('utf-8')
    xml_final_str = head + xml_str + tail
    logger.debug('metadata XML:')
    # logger.debug(xml_final_str.decode())
    return xml_final_str


# def createByteObject(string):
#    string_to_encode = u'\ufeff' + string
#    x = string_to_encode.encode('utf-16be')
#    return ByteStringObject(x)


def _filespec_additional_attachments(
        pdf_filestream, name_arrayobj_cdict, file_dict, file_bin):
    filename = file_dict['filename']
    logger.debug('_filespec_additional_attachments filename=%s', filename)
    mod_date_pdf = _get_pdf_timestamp(file_dict['mod_date'])
    md5sum = hashlib.md5(file_bin).hexdigest()
    md5sum_obj = createStringObject(md5sum)
    params_dict = DictionaryObject({
        NameObject('/CheckSum'): md5sum_obj,
        NameObject('/ModDate'): createStringObject(mod_date_pdf),
        NameObject('/Size'): NameObject(str(len(file_bin))),
        })
    file_entry = DecodedStreamObject()
    file_entry.setData(file_bin)
    file_mimetype = mimetypes.guess_type(filename)[0]
    if not file_mimetype:
        file_mimetype = 'application/octet-stream'
    file_mimetype_insert = '/' + file_mimetype.replace('/', '#2f')
    file_entry.update({
        NameObject("/Type"): NameObject("/EmbeddedFile"),
        NameObject("/Params"): params_dict,
        NameObject("/Subtype"): NameObject(file_mimetype_insert),
        })
    file_entry_obj = pdf_filestream._addObject(file_entry)
    ef_dict = DictionaryObject({
        NameObject("/F"): file_entry_obj,
        })
    fname_obj = createStringObject(filename)
    filespec_dict = DictionaryObject({
        NameObject("/AFRelationship"): NameObject("/Unspecified"),
        NameObject("/Desc"): createStringObject(file_dict.get('desc', '')),
        NameObject("/Type"): NameObject("/Filespec"),
        NameObject("/F"): fname_obj,
        NameObject("/EF"): ef_dict,
        NameObject("/UF"): fname_obj,
        })
    filespec_obj = pdf_filestream._addObject(filespec_dict)
    name_arrayobj_cdict[fname_obj] = filespec_obj

# moved to FacturXPDFWriter
def _facturx_update_metadata_add_attachment(
        pdf_filestream, facturx_xml_str, pdf_metadata, facturx_level,
        output_intents=[], additional_attachments={}):
    '''This method is inspired from the code of the addAttachment()
    method of the PyPDF2 lib'''
    # The entry for the file
    md5sum = hashlib.md5(facturx_xml_str).hexdigest()
    md5sum_obj = createStringObject(md5sum)
    params_dict = DictionaryObject({
        NameObject('/CheckSum'): md5sum_obj,
        NameObject('/ModDate'): createStringObject(_get_pdf_timestamp()),
        NameObject('/Size'): NameObject(str(len(facturx_xml_str))),
        })
    file_entry = DecodedStreamObject()
    file_entry.setData(facturx_xml_str)  # here we integrate the file itself
    file_entry.update({
        NameObject("/Type"): NameObject("/EmbeddedFile"),
        NameObject("/Params"): params_dict,
        # 2F is '/' in hexadecimal
        NameObject("/Subtype"): NameObject("/text#2Fxml"),
        })
    file_entry_obj = pdf_filestream._addObject(file_entry)
    # The Filespec entry
    ef_dict = DictionaryObject({
        NameObject("/F"): file_entry_obj,
        NameObject('/UF'): file_entry_obj,
        })

    fname_obj = createStringObject(FACTURX_FILENAME)
    filespec_dict = DictionaryObject({
        NameObject("/AFRelationship"): NameObject("/Data"),
        NameObject("/Desc"): createStringObject("Factur-X Invoice"),
        NameObject("/Type"): NameObject("/Filespec"),
        NameObject("/F"): fname_obj,
        NameObject("/EF"): ef_dict,
        NameObject("/UF"): fname_obj,
        })
    filespec_obj = pdf_filestream._addObject(filespec_dict)
    name_arrayobj_cdict = {fname_obj: filespec_obj}
    for attach_bin, attach_dict in additional_attachments.items():
        _filespec_additional_attachments(
            pdf_filestream, name_arrayobj_cdict, attach_dict, attach_bin)
    logger.debug('name_arrayobj_cdict=%s', name_arrayobj_cdict)
    name_arrayobj_content_sort = list(
        sorted(name_arrayobj_cdict.items(), key=lambda x: x[0]))
    logger.debug('name_arrayobj_content_sort=%s', name_arrayobj_content_sort)
    name_arrayobj_content_final = []
    af_list = []
    for (fname_obj, filespec_obj) in name_arrayobj_content_sort:
        name_arrayobj_content_final += [fname_obj, filespec_obj]
        af_list.append(filespec_obj)
    embedded_files_names_dict = DictionaryObject({
        NameObject("/Names"): ArrayObject(name_arrayobj_content_final),
        })
    
    # Then create the entry for the root, as it needs a
    # reference to the Filespec
    embedded_files_dict = DictionaryObject({
        NameObject("/EmbeddedFiles"): embedded_files_names_dict,
        })
    res_output_intents = []
    logger.debug('output_intents=%s', output_intents)
    for output_intent_dict, dest_output_profile_dict in output_intents:
        dest_output_profile_obj = pdf_filestream._addObject(
            dest_output_profile_dict)
        # TODO detect if there are no other objects in output_intent_dest_obj
        # than /DestOutputProfile
        output_intent_dict.update({
            NameObject("/DestOutputProfile"): dest_output_profile_obj,
            })
        output_intent_obj = pdf_filestream._addObject(output_intent_dict)
        res_output_intents.append(output_intent_obj)
    
    # Update the root
    metadata_xml_str = _prepare_pdf_metadata_xml(facturx_level, pdf_metadata)
    metadata_file_entry = DecodedStreamObject()
    metadata_file_entry.setData(metadata_xml_str)
    metadata_file_entry.update({
        NameObject('/Subtype'): NameObject('/XML'),
        NameObject('/Type'): NameObject('/Metadata'),
        })
    metadata_obj = pdf_filestream._addObject(metadata_file_entry)
    af_value_obj = pdf_filestream._addObject(ArrayObject(af_list))
    pdf_filestream._root_object.update({
        NameObject("/AF"): af_value_obj,
        NameObject("/Metadata"): metadata_obj,
        NameObject("/Names"): embedded_files_dict,
        # show attachments when opening PDF
        NameObject("/PageMode"): NameObject("/UseAttachments"),
        })
    logger.debug('res_output_intents=%s', res_output_intents)
    if res_output_intents:
        pdf_filestream._root_object.update({
            NameObject("/OutputIntents"): ArrayObject(res_output_intents),
        })
    metadata_txt_dict = _prepare_pdf_metadata_txt(pdf_metadata)
    pdf_filestream.addMetadata(metadata_txt_dict)


def _extract_base_info(facturx_xml_etree):
    namespaces = facturx_xml_etree.nsmap
    date_xpath = facturx_xml_etree.xpath(
        '//rsm:ExchangedDocument/ram:IssueDateTime/udt:DateTimeString',
        namespaces=namespaces)
    date = date_xpath[0].text
    date_dt = datetime.strptime(date, '%Y%m%d')
    inv_num_xpath = facturx_xml_etree.xpath(
        '//rsm:ExchangedDocument/ram:ID', namespaces=namespaces)
    inv_num = inv_num_xpath[0].text
    seller_xpath = facturx_xml_etree.xpath(
        '//ram:ApplicableHeaderTradeAgreement/ram:SellerTradeParty/ram:Name',
        namespaces=namespaces)
    seller = seller_xpath[0].text
    doc_type_xpath = facturx_xml_etree.xpath(
        '//rsm:ExchangedDocument/ram:TypeCode', namespaces=namespaces)
    doc_type = doc_type_xpath[0].text
    base_info = {
        'seller': seller,
        'number': inv_num,
        'date': date_dt,
        'doc_type': doc_type,
        }
    logger.debug('Extraction of base_info: %s', base_info)
    return base_info


def _base_info2pdf_metadata(base_info):
    if base_info['doc_type'] == '381':
        doc_type_name = u'Refund'
    else:
        doc_type_name = u'Invoice'
    date_str = datetime.strftime(base_info['date'], '%Y-%m-%d')
    title = '%s: %s %s' % (
        base_info['seller'], doc_type_name, base_info['number'])
    subject = 'Factur-X %s %s dated %s issued by %s' % (
        doc_type_name, base_info['number'], date_str, base_info['seller'])
    pdf_metadata = {
        'author': base_info['seller'],
        'keywords': u'%s, Factur-X' % doc_type_name,
        'title': title,
        'subject': subject,
        }
    logger.debug('Converted base_info to pdf_metadata: %s', pdf_metadata)
    return pdf_metadata


def _get_original_output_intents(original_pdf):
    output_intents = []
    try:
        pdf_root = original_pdf.trailer['/Root']
        ori_output_intents = pdf_root['/OutputIntents']
        logger.debug('output_intents_list=%s', ori_output_intents)
        for ori_output_intent in ori_output_intents:
            ori_output_intent_dict = ori_output_intent.getObject()
            logger.debug('ori_output_intents_dict=%s', ori_output_intent_dict)
            dest_output_profile_dict =\
                ori_output_intent_dict['/DestOutputProfile'].getObject()
            output_intents.append(
                (ori_output_intent_dict, dest_output_profile_dict))
    except:
        pass
    return output_intents

def _get_pdf_timestamp(date=None):
    if date is None:
        date = datetime.now()
    # example date format: "D:20141006161354+02'00'"
    pdf_date = date.strftime("D:%Y%m%d%H%M%S+00'00'")
    return pdf_date


import sys
from PyQt5.QtWidgets import QApplication

import subprocess
import os
from distutils import spawn
import shutil
from lxml import etree

from .facturx.facturx import FacturX
from .facturx.flavors import xml_flavor
import json
from datetime import datetime as dt

from PyQt5.QtWidgets import (QMainWindow, QAction, QFileDialog, QLineEdit,
                             QLabel, QDockWidget, QSizePolicy, QGridLayout,
                             QScrollArea, QWidget, QMessageBox, QPushButton,
                             QDialog, QComboBox)
from PyQt5.QtGui import QPixmap, QIcon
from PyQt5.QtCore import Qt, QEvent
from PyPDF2 import PdfFileReader
from PyPDF2.generic import IndirectObject

from .populate import PopulateFieldClass


class InvoiceX(QMainWindow):

    def __init__(self):
        super().__init__()

        self.mainWindowLeft = 300
        self.mainWindowTop = 300
        self.mainWindowWidth = 680
        self.mainWindowHeight = 480

        self.fileLoaded = False
        self.dialog = None
        self.initUI()

    def initUI(self):
        """Give structure to window


        Defines basic structure of main window and checks convert command or
        magick based on the platform
        """

        self.statusBar()
        self.setStatusTip('Select a PDF to get started')
        self.set_menu_bar()
        self.set_dockview_fields()
        self.set_center_widget()
        self.set_toolbar()

        self.setGeometry(self.mainWindowLeft, self.mainWindowTop,
                         self.mainWindowWidth, self.mainWindowHeight)
        self.setWindowTitle('Invoice-X')
        self.setWindowIcon(QIcon(os.path.join(os.path.dirname(__file__),
                                              'icons/logo.ico')))
        self.show()

        if not spawn.find_executable('convert'):
            QMessageBox.critical(self, 'Import Error',
                                 "Imagemagick is not installed",
                                 QMessageBox.Ok)
            self.close()

        if sys.platform[:3] == 'win':
            if not spawn.find_executable('magick'):
                QMessageBox.critical(
                    self, 'Import Error',
                    "Imagemagick and GhostScript are not installed",
                    QMessageBox.Ok)
                self.close()

    def set_toolbar(self):
        """Add toolbar options"""
        toolbar = self.addToolBar('File')
        toolbar.addAction(self.openFile)
        toolbar.addAction(self.saveFile)
        toolbar.addAction(self.validateMetadata)
        toolbar.addAction(self.editFields)

    def set_center_widget(self):
        """Add central widget options"""
        self.square = QLabel(self)
        self.square.setAlignment(Qt.AlignCenter)
        self.setCentralWidget(self.square)

    def set_dockview_fields(self):
        """Add Right dock options"""
        self.fields = QDockWidget("Fields", self)
        self.fields.installEventFilter(self)
        self.fieldsQWidget = QWidget()
        self.fieldsScrollArea = QScrollArea()
        self.fieldsScrollArea.setWidgetResizable(True)
        self.fieldsScrollArea.setWidget(self.fieldsQWidget)

        self.layout = QGridLayout()
        self.fieldsQWidget.setLayout(self.layout)

        self.fields.setWidget(self.fieldsScrollArea)
        self.fields.setFloating(False)
        self.fields.setMinimumWidth(360)
        self.fields.setStyleSheet("QWidget { background-color: #AAB2BD}")
        self.addDockWidget(Qt.RightDockWidgetArea, self.fields)

    def set_menu_bar(self):
        """Add Menu bar options"""
        self.exitAct = QAction(QIcon(os.path.join(
            os.path.dirname(__file__), 'icons/exit.png')), 'Exit', self)
        self.exitAct.setShortcut('Ctrl+Q')
        self.exitAct.setStatusTip('Exit application')
        self.exitAct.triggered.connect(self.close)

        self.openFile = QAction(QIcon(os.path.join(
            os.path.dirname(__file__), 'icons/pdf.png')), 'Open', self)
        self.openFile.setShortcut('Ctrl+O')
        self.openFile.setStatusTip('Open new File')
        self.openFile.triggered.connect(self.show_file_dialog)

        self.saveFile = QAction(QIcon(os.path.join(
            os.path.dirname(__file__), 'icons/save.png')), 'Save', self)
        self.saveFile.setShortcut('Ctrl+S')
        self.saveFile.setStatusTip('Save File')
        self.saveFile.triggered.connect(self.save_file_dialog)

        self.saveAsFile = QAction('Save As', self)
        self.saveAsFile.setStatusTip('Save File as a new File')
        self.saveAsFile.triggered.connect(self.show_save_as_dialog)

        self.viewDock = QAction('View Fields', self, checkable=True)
        self.viewDock.setStatusTip('View Fields')
        self.viewDock.setChecked(True)
        self.viewDock.triggered.connect(self._view_dock_field_toggle)

        extractFields = QAction('Extract Fields', self)
        extractFields.setStatusTip('Extract Fields from PDF and add to XML')
        extractFields.triggered.connect(self.extract_fields_from_pdf)

        jsonFormat = QAction('JSON', self)
        jsonFormat.setStatusTip('Export file to JSON')
        jsonFormat.triggered.connect(lambda: self.export_fields('json'))

        xmlFormat = QAction('XML', self)
        xmlFormat.setStatusTip('Export file to XML')
        xmlFormat.triggered.connect(lambda: self.export_fields('xml'))

        ymlFormat = QAction('YML', self)
        ymlFormat.setStatusTip('Export file to YML')
        ymlFormat.triggered.connect(lambda: self.export_fields('yml'))

        self.validateMetadata = QAction(QIcon(os.path.join(
            os.path.dirname(__file__), 'icons/validate.png')),
            'Validate', self)
        self.validateMetadata.setStatusTip('Validate XML')
        self.validateMetadata.triggered.connect(self.validate_xml)

        addMetadata = QAction('Add Metadata', self)
        addMetadata.setStatusTip('Add metadata to PDF')

        self.editFields = QAction(QIcon(
            os.path.join(os.path.dirname(__file__), 'icons/edit.png')),
            'Edit Metadata', self)
        self.editFields.setStatusTip('Edit Metadata in XML')
        self.editFields.triggered.connect(self.edit_fields_dialog)

        documentation = QAction('Documentation', self)
        documentation.setStatusTip('Open Documentation for Invoice-X')
        documentation.triggered.connect(self.documentation_menubar)

        aboutApp = QAction('About', self)
        aboutApp.setStatusTip('Know about Invoice-X')
        aboutApp.triggered.connect(self.about_app_menubar)

        menubar = self.menuBar()
        fileMenu = menubar.addMenu('&File')
        fileMenu.addAction(self.openFile)
        fileMenu.addAction(self.saveFile)
        fileMenu.addAction(self.saveAsFile)
        fileMenu.addAction(self.viewDock)
        fileMenu.addAction(self.exitAct)

        commandMenu = menubar.addMenu('&Command')

        exportMetadata = commandMenu.addMenu('&Export Metadata')
        exportMetadata.addAction(jsonFormat)
        exportMetadata.addAction(xmlFormat)
        exportMetadata.addAction(ymlFormat)

        commandMenu.addAction(self.validateMetadata)
        commandMenu.addAction(self.editFields)
        commandMenu.addAction(addMetadata)
        commandMenu.addAction(extractFields)

        helpMenu = menubar.addMenu('&Help')
        helpMenu.addAction(documentation)
        helpMenu.addAction(aboutApp)

    def _view_dock_field_toggle(self, state):
        """Toggle to view Dock"""
        if state:
            self.fields.show()
        else:
            self.fields.hide()

    def validate_xml(self):
        """Validate XML using is_valid()"""
        try:
            if self.factx.is_valid():
                QMessageBox.information(self, 'Valid XML',
                                        "The XML is Valid",
                                        QMessageBox.Ok)
            else:
                QMessageBox.critical(self, 'Invalid XML',
                                     "The XML is invalid",
                                     QMessageBox.Ok)
        except AttributeError:
            QMessageBox.critical(self, 'File Not Found',
                                 "Load a PDF first",
                                 QMessageBox.Ok)

    def set_pdf_preview(self):
        """View preview of PDF opened

        Uses imagemagick and ghostscript to create image from pdf
        """
        if not os.path.exists('.load'):
            os.mkdir('.load')
        if sys.platform[:3] == 'win':
            convert = ['magick', self.fileName[0],
                       '-flatten', '.load/preview.jpg']
        else:
            convert = ['convert', '-verbose', '-density', '150', '-trim',
                       self.fileName[0], '-quality', '100', '-flatten',
                       '-sharpen', '0x1.0', '.load/preview.jpg']
        subprocess.call(convert)
        self.pdfPreviewImage = '.load/preview.jpg'
        self.fileLoaded = True
        self.square.setPixmap(QPixmap(self.pdfPreviewImage).scaled(
            self.square.size().width(), self.square.size().height(),
            Qt.KeepAspectRatio, Qt.SmoothTransformation))

    def edit_fields_dialog(self):
        """function to edit fields"""
        try:
            self.dialog = EditFieldsClass(self, self.factx,
                                          self.fieldsDict,
                                          self.metadata_field)
            self.dialog.installEventFilter(self)
            # self.dialog.show()
        except AttributeError:
            QMessageBox.critical(self, 'File Not Found',
                                 "Load a PDF first",
                                 QMessageBox.Ok)

    def update_dock_fields(self):
        """Load Fields from the attached XML"""
        self.factx.write_json('.load/output.json')
        with open('.load/output.json') as jsonFile:
            self.fieldsDict = json.load(jsonFile)
        os.remove('.load/output.json')
        # print(self.fieldsDict)

        i = 0

        self.metadata_field = {
            'amount_tax': 'Amount Tax',
            'amount_total': 'Amount Total',
            'amount_untaxed': 'Amount Untaxed',
            'buyer': 'Buyer',
            'currency': 'Currency',
            'date': 'Date',
            'date_due': 'Date Due',
            'invoice_number': 'Invoice Number',
            'name': 'Name',
            'notes': 'Notes',
            'seller': 'Seller',
            'type': 'Type',
            'version': 'Version'
        }

        for key in sorted(self.fieldsDict):
            i += 1
            try:
                self.factx[key]
            except IndexError:
                self.fieldsDict[key] = "Field Not Specified"
            except TypeError:
                pass
            fieldKey = QLabel(self.metadata_field[key] + ": ")
            if self.fieldsDict[key] is None:
                fieldValue = QLabel("NA")
            else:
                if key[:4] == "date" and \
                        self.fieldsDict[key] != "Field Not Specified":
                    self.fieldsDict[key] = self.fieldsDict[key][:4] \
                        + "/" + self.fieldsDict[key][4:6] \
                        + "/" + self.fieldsDict[key][6:8]
                if self.fieldsDict[key] == "Field Not Specified":
                    fieldValue = QLabel(self.fieldsDict[key])
                    fieldValue.setStyleSheet("QLabel { color: #666666}")
                else:
                    fieldValue = QLabel(self.fieldsDict[key])
            self.layout.addWidget(fieldKey, i, 0)
            self.layout.addWidget(fieldValue, i, 1)

    def show_file_dialog(self):
        """OPen File Dialog to select PDF to be opened"""
        self.fileName = QFileDialog.getOpenFileName(self, 'Open file',
                                                    os.path.expanduser("~"),
                                                    "pdf (*.pdf)")
        self.load_pdf_file()

    def load_pdf_file(self):
        """Load PDF and find standard

        If no attached XML is found, then show dialog to select Standard
        """
        if self.fileName[0]:
            if self.check_xml_for_pdf() is None:
                self.standard = None
                self.level = None
                self._choose_standard_level()
                if self.standard is not None:
                    self.factx = FacturX(self.fileName[0],
                                         self.standard,
                                         self.level)
            else:
                self.factx = FacturX(self.fileName[0])
            if hasattr(self, 'factx'):
                self.set_pdf_preview()
                self.update_dock_fields()
                self.setStatusTip("PDF is Ready")

    def _choose_standard_level(self):
        self.chooseStandardDialog = QDialog()
        layout = QGridLayout()

        noXMLLabel = QLabel("No XML found", self)
        layout.addWidget(noXMLLabel, 0, 0)

        chooseStandardLabel = QLabel("Standard", self)
        chooseStandardCombo = QComboBox(self)
        chooseStandardCombo.addItem("Factur-X")
        chooseStandardCombo.addItem("Zugferd")
        chooseStandardCombo.addItem("UBL")
        chooseStandardCombo.model().item(2).setEnabled(False)
        chooseStandardCombo.activated[str].connect(self._on_select_level)

        chooseLevelLabel = QLabel("Level", self)
        self.chooseLevelCombo = QComboBox(self)
        self.chooseLevelCombo.addItem("Minimum")
        self.chooseLevelCombo.addItem("Basic WL")
        self.chooseLevelCombo.addItem("Basic")
        self.chooseLevelCombo.addItem("EN16931")
        self.chooseLevelCombo.activated[str].connect(self._set_level)

        applyStandard = QPushButton("Apply")
        applyStandard.clicked.connect(self._set_standard_level)
        discardStandard = QPushButton("Cancel")
        discardStandard.clicked.connect(self._discard_standard_level)

        layout.addWidget(chooseStandardLabel, 1, 0)
        layout.addWidget(chooseStandardCombo, 1, 1)
        layout.addWidget(chooseLevelLabel, 2, 0)
        layout.addWidget(self.chooseLevelCombo, 2, 1)
        layout.addWidget(discardStandard, 3, 0)
        layout.addWidget(applyStandard, 3, 1)

        self.chooseStandardDialog.setLayout(layout)
        self.chooseStandardDialog.setWindowTitle("Choose Standard")
        self.chooseStandardDialog.setWindowModality(Qt.ApplicationModal)
        self.chooseStandardDialog.exec_()

    def _set_standard_level(self):
        if hasattr(self, 'standard_temp'):
            self.standard = self.standard_temp
        else:
            self.standard = 'factur-x'

        if hasattr(self, 'level_temp'):
            self.level = self.level_temp
        else:
            self.level = 'minimum'
        self.chooseStandardDialog.close()

    def _discard_standard_level(self):
        self.chooseStandardDialog.close()

    def _on_select_level(self, text):
        if text == "Factur-X":
            self.chooseLevelCombo.clear()
            self.chooseLevelCombo.addItem("Minimum")
            self.chooseLevelCombo.addItem("Basic WL")
            self.chooseLevelCombo.addItem("Basic")
            self.chooseLevelCombo.addItem("EN16931")
        elif text == "Zugferd":
            self.chooseLevelCombo.clear()
            self.chooseLevelCombo.addItem("Basic")
            self.chooseLevelCombo.addItem("Comfort")
        elif text == "UBL":
            self.chooseLevelCombo.clear()
            self.chooseLevelCombo.addItem("UBL 2.0")
            self.chooseLevelCombo.addItem("UBL 2.1")

        standard_dict = {
            'Factur-X': ['factur-x', 'minimum'],
            'Zugferd': ['zugferd', 'basic'],
        }
        self.standard_temp = standard_dict[text][0]
        self.level_temp = standard_dict[text][1]

    def _set_level(self, text):
        level_dict = {
            'Minimum': 'minimum',
            'Basic WL': 'basicwl',
            'Basic': 'basic',
            'EN16931': 'en16931',
            'Comfort': 'comfort'
        }
        self.level_temp = level_dict[text]

    def check_xml_for_pdf(self):
        """Look for XML in PDF"""
        pdf = PdfFileReader(self.fileName[0])
        pdf_root = pdf.trailer['/Root']
        if '/Names' not in pdf_root or '/EmbeddedFiles' not in \
                pdf_root['/Names']:
            return None

        for file in pdf_root['/Names']['/EmbeddedFiles']['/Names']:
            if isinstance(file, IndirectObject):
                obj = file.getObject()
                if obj['/F'] in xml_flavor.valid_xmp_filenames():
                    xml_root = etree.fromstring(obj['/EF']['/F'].getData())
                    xml_content = xml_root
        return xml_content

    def save_file_dialog(self):
        """Open dialog to select location"""
        if self.fileLoaded:
            if self.confirm_save_dialog():
                try:
                    self.factx.write_pdf(self.fileName[0])
                except TypeError:
                    QMessageBox.critical(self, 'Type Error',
                                         "Some field value(s) are invalid",
                                         QMessageBox.Ok)
        else:
            QMessageBox.critical(self, 'File Not Found',
                                 "Load a PDF first",
                                 QMessageBox.Ok)

    def confirm_save_dialog(self):
        """Prompt to verify saving of PDF with XML"""
        reply = QMessageBox.question(
            self, 'Message', "Do you want to save? This cannot be undone",
            QMessageBox.Yes | QMessageBox.No, QMessageBox.No)

        if reply == QMessageBox.Yes:
            return True
        else:
            return False

    def show_save_as_dialog(self):
        """Open dialog to select location"""
        if self.fileLoaded:
            try:
                self.saveFileName = QFileDialog.getSaveFileName(
                    self, 'Save file', os.path.expanduser("~"), "pdf (*.pdf)")
                if self.saveFileName[0]:
                    if self.saveFileName[0].endswith('.pdf'):
                        fileName = self.saveFileName[0]
                    else:
                        fileName = self.saveFileName[0] + '.pdf'
                    self.factx.write_pdf(fileName)
            except TypeError:
                QMessageBox.critical(self, 'Type Error',
                                     "Some field value(s) are not valid",
                                     QMessageBox.Ok)
        else:
            QMessageBox.critical(self, 'File Not Found',
                                 "Load a PDF first",
                                 QMessageBox.Ok)

    def extract_fields_from_pdf(self):
        """Integration with invoice2data"""
        if self.fileLoaded:
            self.populate = PopulateFieldClass(self, self.factx,
                                               self.fieldsDict,
                                               self.metadata_field)
        else:
            QMessageBox.critical(self, 'File Not Found',
                                 "Load a PDF first",
                                 QMessageBox.Ok)

    def documentation_menubar(self):
        pass

    def about_app_menubar(self):
        pass

    def export_fields(self, outputformat):
        """Export metadata in XML, JSON, YAML format"""
        if self.fileLoaded:
            self.exportFileName = QFileDialog.getSaveFileName(
                self, 'Export file',
                os.path.expanduser("~") +
                '/output.%s' % outputformat,
                "%s (*.%s)" % (outputformat, outputformat))
            if self.exportFileName[0]:
                if outputformat is "json":
                    self.pdf_write_json(self.exportFileName[0])
                elif outputformat is "xml":
                    self.pdf_write_xml(self.exportFileName[0])
                elif outputformat is "yml":
                    self.pdf_write_yml(self.exportFileName[0])
        else:
            QMessageBox.critical(self, 'File Not Found',
                                 "Load a PDF first",
                                 QMessageBox.Ok)

    def pdf_write_json(self, fileName):
        self.factx.write_json(fileName)

    def pdf_write_xml(self, fileName):
        self.factx.write_xml(fileName)

    def pdf_write_yml(self, fileName):
        self.factx.write_yml(fileName)

    def resizeEvent(self, event):
        """Fix for pixelated images from imagemagick"""
        if self.fileLoaded:
            self.square.setPixmap(QPixmap(self.pdfPreviewImage).scaled(
                self.square.size().width(), self.square.size().height(),
                Qt.KeepAspectRatio, Qt.SmoothTransformation))
            self.square.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Ignored)
            QMainWindow.resizeEvent(self, event)

    def eventFilter(self, source, event):
        if event.type() == QEvent.Close and source is self.fields:
            self.viewDock.setChecked(False)
        return QMainWindow.eventFilter(self, source, event)

    def closeEvent(self, event):
        if os.path.isdir('.load'):
            shutil.rmtree('.load/')


class EditFieldsClass(QWidget, object):
    def __init__(self, invx, factx, fieldsDict, metadataDict):
        super().__init__()
        self.fDict = fieldsDict
        self.mDict = metadataDict
        self.factx = factx
        self.invx = invx
        self.initUI()

    def initUI(self):
        """Set up dialog to edit Fields"""
        layout = QGridLayout()
        i = 0
        self.fieldsKeyList = []
        self.fieldsValueList = []
        for key in sorted(self.fDict):
            i += 1
            fKey = QLabel(self.mDict[key])
            fValue = QLineEdit()
            fValue.setText(self.fDict[key])
            if key[:4] == 'date':
                fValue.setPlaceholderText("YYYY/MM/DD")
            if self.fDict[key] == "Field Not Specified":
                fValue.setEnabled(False)
            else:
                self.fieldsKeyList.append(key)
                self.fieldsValueList.append(fValue)
            layout.addWidget(fKey, i, 0)
            layout.addWidget(fValue, i, 1)

        i = i + 1
        saveButton = QPushButton('Apply')
        saveButton.clicked.connect(self.update_fields_and_dock)
        discardButton = QPushButton('Discard')
        discardButton.clicked.connect(self.discard_fields)
        layout.addWidget(discardButton, i, 0)
        layout.addWidget(saveButton, i, 1)

        self.setLayout(layout)
        self.move(300, 150)
        self.setWindowTitle('Edit Fields')
        self.setWindowIcon(QIcon('icons/logo.png'))
        self.show()

    def update_fields_and_dock(self):
        """Update fields and update dock"""
        try:
            for key, value in zip(self.fieldsKeyList, self.fieldsValueList):
                if key[:4] != "date":
                    self.factx[key] = value.text()
                else:
                    self.factx[key] = dt.strptime(value.text(), '%Y/%m/%d')
            self.invx.update_dock_fields()
            self.close()
        except ValueError:
            QMessageBox.critical(self, 'Invalid Field Value',
                                 "Invalid Field Value(s)",
                                 QMessageBox.Ok)

    def discard_fields(self):
        self.close()


def main():
    app = QApplication(sys.argv)
    InvoiceX()
    sys.exit(app.exec_())


if __name__ == '__main__':
    main()



import os
import unittest
from facturx.facturx import *
from lxml import etree


class TestReading(unittest.TestCase):
    def discover_files(self):
        self.test_files_dir = os.path.join(os.path.dirname(__file__), 'sample_invoices')
        self.test_files = os.listdir(self.test_files_dir)

    def test_from_file(self):
        self.discover_files()
        for file in self.test_files:
            file_path = os.path.join(self.test_files_dir, file)
            FacturX(file_path)

    # returning file path for a specific file in 'sample_invoices'
    def find_file(self, file_name):
        self.discover_files()
        for file in self.test_files:
            if file == file_name:
                file_path = os.path.join(self.test_files_dir, file)
                return file_path

    # def test_input_error(self):
    #     with self.assertRaises(TypeError) as context:
    #         FacturX('non-existant.pdf')

    def test_file_without_embedded_data(self):
        file_path = self.find_file('no_embedded_data.pdf')
        self.assertEqual(FacturX(file_path)._xml_from_file(file_path), None)

    def test_file_embedded_data(self, file_name='embedded_data.pdf'):
        file_path = self.find_file(file_name)
        self.assertTrue(FacturX(file_path)._xml_from_file(file_path) is not None, "The PDF file has no embedded file")

    def test_write_pdf(self):
        file_path = self.find_file('no_embedded_data.pdf')
        factx = FacturX(file_path)
        test_file_path = os.path.join(self.test_files_dir, 'test.pdf')

        # checking if pdf file is made
        factx.write_pdf(test_file_path)
        self.assertTrue(os.path.isfile(test_file_path))
        self.discover_files()

        # checking that xml is embedded
        self.assertTrue(self.test_file_embedded_data(file_name='test.pdf') is None)

        os.remove(test_file_path)

    def test_write_xml(self):
        compare_file_dir = os.path.join(os.path.dirname(__file__), 'compare')
        expected_file_path = os.path.join(compare_file_dir, 'no_embedded_data.xml')
        test_file_path = os.path.join(compare_file_dir, 'test.xml')

        factx = FacturX(self.find_file('no_embedded_data.pdf'))
        factx.write_xml(test_file_path)
        self.assertTrue(os.path.isfile(test_file_path))

        with open(expected_file_path, 'r') as expected_file, open(test_file_path, 'r') as test_file:
            parser = etree.XMLParser(remove_blank_text=True)
            expected_file_root = etree.XML(expected_file.read(), parser)
            expected_file_str = etree.tostring(expected_file_root)

            test_file_root = etree.XML(test_file.read(), parser)
            test_file_str = etree.tostring(test_file_root)

        self.assertTrue(expected_file_str == test_file_str, "Files don't match")
        os.remove(test_file_path)


def main():
    unittest.main()


if __name__ == '__main__':
    main()



from PyQt5.QtWidgets import (QWidget, QPushButton, QGridLayout, QFileDialog,
                             QLabel, QLineEdit, QMessageBox, QCheckBox)
from PyQt5.QtGui import QIcon, QFont
import configparser
from invoice2data.main import main, create_parser
import json


class PopulateFieldClass(QWidget):
    def __init__(self, gui, factx, fieldsDict, metadataDict):
        super().__init__()
        self.gui = gui
        self.fDict = fieldsDict
        self.mDict = metadataDict
        self.factx = factx
        self.customTemplateFolderName = None
        self.init_ui()

    def init_ui(self):
        """Setup layout to add default field values"""
        layout = QGridLayout()
        i = 3
        customTemplate = QPushButton('Custom Template')
        customTemplate.clicked.connect(self.customTemplateDialog)
        layout.addWidget(customTemplate, 0, 0)

        self.excludeDefaultFolder = QCheckBox(
            'Exclude default template folder', self)
        layout.addWidget(self.excludeDefaultFolder, 1, 1)

        self.customTemplateLineEdit = QLineEdit()
        self.customTemplateLineEdit.setDisabled(True)
        layout.addWidget(self.customTemplateLineEdit, 0, 1)

        labelBlank = QLabel(" ")
        layout.addWidget(labelBlank, 1, 0)

        labelTitle = QLabel("Add Default Values")
        labelTitle.setFont(QFont("Add Default Values", weight=QFont.Bold))
        layout.addWidget(labelTitle, 2, 0)
        self.fieldsKeyList = []
        self.fieldsValueList = []
        for key in sorted(self.fDict):
            i += 1
            fKey = QLabel(self.mDict[key])
            fValue = QLineEdit()
            fValue.setText("")
            if self.fDict[key] == "Field Not Specified" or key[:4] == "date":
                fValue.setEnabled(False)
            else:
                self.fieldsKeyList.append(key)
                self.fieldsValueList.append(fValue)
            layout.addWidget(fKey, i, 0)
            layout.addWidget(fValue, i, 1)

        i += 1
        saveButton = QPushButton('Apply')
        saveButton.clicked.connect(self.call_invoice2data)
        resetButton = QPushButton('Discard')
        resetButton.clicked.connect(self.resetLabel)
        layout.addWidget(resetButton, i, 0)
        layout.addWidget(saveButton, i, 1)

        self.setLayout(layout)
        self.move(300, 150)
        self.setWindowTitle('Edit Fields')
        self.setWindowIcon(QIcon('icons/logo.png'))
        self.show()

    def customTemplateDialog(self):
        """Dialog to select location of custom Template Folder"""
        self.customTemplateFolderName = QFileDialog.getExistingDirectory(
            self, "Select Custom Template Folder")
        if self.customTemplateFolderName:
            self.customTemplateLineEdit.setText(self.customTemplateFolderName)

    def call_invoice2data(self):
        """Invoke invoice2data"""
        if self.excludeDefaultFolder.isChecked() and \
                self.customTemplateFolderName is None:
            QMessageBox.critical(self, 'Error',
                                 "Select a custom template first",
                                 QMessageBox.Ok)
        else:
            config = configparser.ConfigParser()
            config['CUSTOM'] = {}
            for key, value in zip(self.fieldsKeyList, self.fieldsValueList):
                if value.text() is not "":
                    config['CUSTOM'][key] = value.text()
            with open('.load/default.cfg', 'w') as configfile:
                config.write(configfile)
            populate_using_invoice2data(self,
                                        self.excludeDefaultFolder.isChecked(),
                                        self.customTemplateFolderName,
                                        self.fDict,
                                        self.gui,
                                        self.factx)
            self.close()

    def resetLabel(self):
        self.close()


class populate_using_invoice2data(object):
    def __init__(self, popfield, excludeDefaultFolder, templateFolder,
                 fieldDict, gui, factx):
        self.parser = create_parser()
        self.fieldValueDict = fieldDict
        self.gui = gui
        self.factx = factx
        self.popfield = popfield
        self.outputFile = '.load/invoice2data_output.json'
        self.parseList = ['--output-format',
                          'json',
                          '--output-name',
                          self.outputFile,
                          '--exclude-built-in-templates',
                          '--template-folder',
                          templateFolder,
                          self.gui.fileName[0]]
        if not excludeDefaultFolder:
            self.parseList.remove('--exclude-built-in-templates')
        if templateFolder is None:
            self.parseList.remove('--template-folder')
            self.parseList.remove(templateFolder)

        main(self.parser.parse_args(self.parseList))

        self.set_values()

    def set_values(self):
        """Set field values using invoice2data"""
        templateerror = False
        fieldMatchDict = {
            'seller': 'issuer',
            'amount_total': 'amount'
        }
        try:
            with open(self.outputFile, 'r') as json_file:
                invoiceFieldDict = json.load(json_file)
                config = configparser.RawConfigParser()
                config.read('.load/default.cfg')
                for key, value in config['CUSTOM'].items():
                    if key in fieldMatchDict:
                        invoiceFieldDict[0][fieldMatchDict[key]] = value

                    invoiceFieldDict[0][key] = value
                for key in self.fieldValueDict:
                    if key in invoiceFieldDict[0]:
                        self.fieldValueDict[key] = invoiceFieldDict[0][key]
                    if key in fieldMatchDict:
                            self.fieldValueDict[key] = \
                                invoiceFieldDict[0][fieldMatchDict[key]]
        except IndexError:
            templateerror = True
            QMessageBox.critical(self.popfield, 'Template Error',
                                 "No matching template found",
                                 QMessageBox.Ok)

        if not templateerror:
            for key, value in self.fieldValueDict.items():
                    try:
                        if key[:4] != "date":
                            if value is None:
                                self.factx[key] = "NA"
                            else:
                                self.factx[key] = str(value)
                        else:
                            pass
                    except IndexError:
                        pass

            self.gui.update_dock_fields()



from ..invoicex import InvoiceX
import unittest
import os
from PyQt5.QtWidgets import QApplication
import sys
import shutil


class TestMainWindow(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        app = QApplication(sys.argv)
        cls.widget = InvoiceX()
        test_file = os.path.join(os.path.dirname(__file__),
                                 'sample_invoice/Facture_FR_BASIC.pdf')
        cls.widget.fileName = [test_file]
        cls.widget.load_pdf_file()

    # @classmethod
    # def tearDown(cls):
    #     sys.exit()

    def test_pdf_preview(self):
        self.assertTrue(os.path.isfile('.load/preview.jpg'),
                        "Image file of pdf was not generated")

    def test_save_as(self):
        if not os.path.isdir('temp_test_file'):
            os.mkdir('temp_test_file')
        file_path = 'temp_test_file/output.xml'
        self.widget.pdf_write_xml(file_path)
        self.assertTrue(os.path.isfile(file_path))
        shutil.rmtree('temp_test_file')


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages
from os import path


setup(
    name='invoicex-gui',
    version='0.0.9',
    author='Harshit Joshi',
    author_email='harshit.joshi@outlook.com',
    url='https://github.com/invoice-x/invoicex-gui',
    description='Graphical User Interface for factur-x library with basic functionalities',
    license="GNU General Public License (GPL)",
    long_description=open(path.join(path.dirname(__file__), 'README.rst')).read(),
    package_data = {
        'invoicex.facturx.flavors': [
            '*.yml',
            'factur-x/xml/*.xml',
            'factur-x/xmp/*.xmp',
            'factur-x/xsd/*.xsd',
            'zugferd/xml/*.xml',
            'zugferd/xmp/*.xmp',
            'zugferd/xsd/*.xsd',
           ],
        'invoicex':['icons/*.png']
        },
    packages=find_packages(),
    install_requires=[
        r.strip() for r in open(
            path.join(path.dirname(__file__), 'requirements.txt')
                ).read().splitlines() if not r.startswith('#')
        ],
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'invoicex-gui = invoicex.invoicex:main',
        ]
    }
)